"use strict";

const HTMLMinifier = require("html-minifier").minify;
const StrictHTML = require("stricthtml");

const zc_minify_js = require("./zc_minify_js");
const zc_minify_css = require("./zc_minify_css");

const call_onload = require("../utils/call_onload");

const zc_minify_html = (code, settings, ref) => {
  // Don't call onloadfile here, as it might not be a file (e.g. nested code)
  return new Promise((resolve, reject) => {
    // Minifier doesn't support async functions, so do manually
    // Also, minifier doesn't give position on error
    StrictHTML.parse(code, (err, root) => {
      if (err) {
        reject(err);
        return;
      }

      let queue = root.children.filter(n => n.type == "NT_ELEMENT");
      let to_min = [];

      let node;
      while (node = queue.shift()) {
        if (node.tagName == "script") {
          to_min.push([
            call_onload(settings.types.js.onload, node.children[0].value, ref, false)
              .then(code => zc_minify_js(code, settings, ref)),
            node.children[0],
          ]);

        } else if (node.tagName == "style") {
          to_min.push([
            call_onload(settings.types.css.onload, node.children[0].value, ref, false)
              .then(code => zc_minify_css(code, settings, ref)),
            node.children[0]
          ]);

        } else {
          node.attrs.forEach(a => {
            if (a.name == "style") {
              to_min.push([
                call_onload(settings.types.css.onload, a.value, ref, true)
                  .then(code => zc_minify_css(`a{${code}}`, settings, ref))
                  .then(p => p.slice(2, -1)),
                a
              ]);

            } else if (settings.types.html.eventAttributes.some(reg => reg.test(a.name))) {
              to_min.push([
                call_onload(settings.types.js.onload, a.value, ref, true)
                  .then(code => zc_minify_js(code, settings, ref)),
                a
              ]);
            }
          });

          node.children.forEach(c => {
            if (c.type == "NT_ELEMENT") {
              queue.push(c);
            }
          });
        }
      }

      Promise.all(to_min.map(m => {
        return Promise.resolve(m[0])
          .then(min => {
            m[1].value = min;
          })
          .catch(err => {
            if (err.pos != null) {
              err.pos = m[1].pos + err.pos;
            } else {
              err.pos = m[1].pos;
            }
            return Promise.reject(err);
          });
      }))
        .then(() => {
          if (!settings.types.html.minify) {
            resolve(code);
            return;
          }

          // HTMLMinifier does not set .pos on error
          resolve(HTMLMinifier(StrictHTML.stringify(root), {
            customEventAttributes: [],
            minifyCSS: false,
            minifyJS: false,

            collapseBooleanAttributes: settings.types.html.minify.collapseBooleanAttributes,
            collapseInlineTagWhitespace: settings.types.html.minify.collapseInlineTagWhitespace,
            collapseWhitespace: settings.types.html.minify.collapseWhitespace,
            decodeEntities: settings.types.html.minify.decodeEntities,
            includeAutoGeneratedTags: settings.types.html.minify.includeAutoGeneratedTags,
            removeAttributeQuotes: settings.types.html.minify.removeAttributeQuotes,
            removeComments: settings.types.html.minify.removeComments,
            removeEmptyAttributes: settings.types.html.minify.removeEmptyAttributes,
            removeOptionalTags: settings.types.html.minify.removeOptionalTags,
            removeRedundantAttributes: settings.types.html.minify.removeRedundantAttributes,
            removeScriptTypeAttributes: settings.types.html.minify.removeScriptTypeAttributes,
            removeStyleLinkTypeAttributes: settings.types.html.minify.removeStyleLinkTypeAttributes,
            removeTagWhitespace: settings.types.html.minify.removeTagWhitespace,
            sortAttributes: settings.types.html.minify.sortAttributes,
            sortClassName: settings.types.html.minify.sortClassName,
            trimCustomFragments: settings.types.html.minify.trimCustomFragments,
          })
            .replace(/&zcsp;/g, " ")
            .replace(/&zclf;/g, "\n")
            .replace(/&zctab;/g, "\t"));
        })
        .catch(reject);
    });
  });
};

module.exports = zc_minify_html;
